from _pytest import main as pytest_main
from find_dependencies.dependency_finder import DependencyFinder, run_tests


def pytest_addoption(parser):
    group = parser.getgroup("find-dependencies")
    group.addoption(
        "--find-dependencies",
        action="store_true",
        dest="find_dependencies",
        help="""Find dependencies between tests""",
    )
    group.addoption(
        "--reversed-first",
        action="store_true",
        dest="reversed_first",
        help="""Run the test in reverse direction first""",
    )
    group.addoption(
        "--find-dependencies-internal",
        action="store_true",
        dest="find_dependencies_internal",
        help="""For internal use only""",
    )


def pytest_runtestloop(session):
    if session.config.getoption("find_dependencies_internal"):
        return run_tests(session)
    if not session.config.getoption("find_dependencies"):
        return pytest_main.pytest_runtestloop(session)

    if len(session.items) == 1:
        print("Only one test collected: ignoring option --find-dependencies")
        return pytest_main.pytest_runtestloop(session)

    if (session.testsfailed and
            not session.config.option.continue_on_collection_errors):
        raise session.Interrupted(
            "%d errors during collection" % session.testsfailed)

    if session.config.option.collectonly:
        return True

    DependencyFinder(session).find_dependencies()
    return True



import pytest

pytest_plugins = ["pytester"]


@pytest.fixture
def test_path(testdir):
    testdir.tmpdir.join("pytest.ini").write(
        "[pytest]\n" "console_output_style = classic"
    )
    yield testdir


def test_no_checks_if_not_configured(test_path):
    test_path.makepyfile(
        test_one="""
        def test_a(): pass
        def test_b(): pass
        def test_c(): pass
        """
    )
    result = test_path.runpytest("-v", "-p", "no:randomly")
    result.assert_outcomes(passed=3, failed=0)
    result.stdout.fnmatch_lines([
        "test_one.py::test_a PASSED",
        "test_one.py::test_b PASSED",
        "test_one.py::test_c PASSED"
    ])


def test_no_checks_for_single_test(test_path):
    test_path.makepyfile(
        test_one="""
        def test_a(): pass
        """
    )
    result = test_path.runpytest("-v", "--find-dependencies")
    result.assert_outcomes(passed=1, failed=0)
    result.stdout.fnmatch_lines([
        "Only one test collected: ignoring option --find-dependencies",
        "test_one.py::test_a PASSED",
    ])


def test_no_checks_if_collection_failed(test_path):
    test_path.makepyfile(
        test_one="""
        def test_a(): pass
        def test_b(): pass
        """
    )
    test_path.makepyfile(
        conftest="""
        def pytest_collection(session):
            session.perform_collect()
            session.testsfailed = 1
        """
    )
    result = test_path.runpytest("--find-dependencies")
    result.assert_outcomes(passed=0, failed=0)
    result.stdout.fnmatch_lines([
        "*Interrupted: 1 errors during collection*",
        "*no tests ran*"
    ])


def test_no_dependencies(test_path):
    test_path.makepyfile(
        test_one="""
        def test_a(): pass
        def test_b(): assert False
        def test_c(): pass
        """
    )

    result = test_path.runpytest("--find-dependencies")
    result.stdout.fnmatch_lines([
        "The following tests are always failing and are "
        "excluded from the analysis:",
        "  test_one.py::test_b",
        "No dependent tests found."
    ])


def test_single_dependency_collect_only(test_path):
    test_path.makepyfile(
        test_one="""
        flag = True
        def test_a(): pass
        def test_b(): assert flag
        def test_c(): global flag; flag = False
        """
    )

    result = test_path.runpytest("--find-dependencies", "--collect-only",
                                 "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "collected 3 items",
        "<Module test_one.py>",
        "  <Function test_a>",
        "  <Function test_b>",
        "  <Function test_c>"
    ])


def test_single_dependency_last_index(test_path):
    test_path.makepyfile(
        test_one="""
        flag = True
        def test_a(): pass
        def test_b(): assert flag
        def test_c(): global flag; flag = False
        """
    )

    result = test_path.runpytest("--find-dependencies", "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 3 tests.",
        "Executed 7 tests in 3 test runs.",
        "Dependent tests:",
        "  test_one.py::test_b depends on test_one.py::test_c"
    ])


def test_single_dependency_first_index(test_path):
    test_path.makepyfile(
        test_one="""
        flag = True
        def test_a(): global flag; flag = False
        def test_b(): assert flag
        def test_c(): pass
        """
    )

    result = test_path.runpytest("--find-dependencies", "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 3 tests.",
        "Executed 6 tests in 2 test runs.",
        "Dependent tests:",
        "  test_one.py::test_b depends on test_one.py::test_a"
    ])


def test_single_dependency1(test_path):
    test_path.makepyfile(
        test_one="""
        flag = True
        def test_a(): pass
        def test_b(): assert flag
        def test_c(): global flag; flag = False
        def test_d(): pass
        """
    )

    result = test_path.runpytest("--find-dependencies", "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 4 tests.",
        "Executed 11 tests in 4 test runs.",
        "Dependent tests:",
        "  test_one.py::test_b depends on test_one.py::test_c"
    ])


def test_single_reversed_first(test_path):
    test_path.makepyfile(
        test_one="""
        flag = True
        def test_a(): pass
        def test_b(): assert flag
        def test_c(): global flag; flag = False
        def test_d(): pass
        """
    )

    result = test_path.runpytest("--find-dependencies", "--reversed-first",
                                 "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 4 tests.",
        "Executed 10 tests in 3 test runs.",
        "Dependent tests:",
        "  test_one.py::test_b depends on test_one.py::test_c"
    ])


def test_single_dependency2(test_path):
    test_path.makepyfile(
        test_one="""
        flag = True
        def test_a(): pass
        def test_b(): assert flag
        def test_c(): pass
        def test_d(): pass
        def test_e(): pass
        def test_f(): pass
        def test_g(): global flag; flag = False
        def test_h(): pass
        def test_i(): pass
        """
    )

    result = test_path.runpytest("--find-dependencies", "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 9 tests.",
        "Executed 27 tests in 6 test runs.",
        "Dependent tests:",
        "  test_one.py::test_b depends on test_one.py::test_g"
    ])


def test_single_dependency3(test_path):
    test_path.makepyfile(
        test_one="""
        flag = True
        def test_a(): pass
        def test_b(): global flag; flag = False
        def test_c(): pass
        def test_d(): pass
        def test_e(): assert flag
        def test_f(): pass
        def test_g(): pass
        """
    )

    result = test_path.runpytest("--find-dependencies", "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 7 tests.",
        "Executed 19 tests in 4 test runs.",
        "Dependent tests:",
        "  test_one.py::test_e depends on test_one.py::test_b"
    ])


def test_single_dependency1_with_randomly(test_path):
    test_path.makepyfile(
        test_one="""
        flag = True
        def test_a(): pass
        def test_b(): assert flag
        def test_c(): global flag; flag = False
        def test_d(): pass
        """
    )

    result = test_path.runpytest("--find-dependencies")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 4 tests.",
        "Dependent tests:",
        "  test_one.py::test_b depends on test_one.py::test_c"
    ])


def test_single_dependency2_with_randomly(test_path):
    test_path.makepyfile(
        test_one="""
        flag = True
        def test_a(): pass
        def test_b(): assert flag
        def test_c(): pass
        def test_d(): pass
        def test_e(): pass
        def test_f(): pass
        def test_g(): global flag; flag = False
        def test_h(): pass
        def test_i(): pass
        """
    )

    result = test_path.runpytest("--find-dependencies", "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 9 tests.",
        "Dependent tests:",
        "  test_one.py::test_b depends on test_one.py::test_g"
    ])


def test_single_dependency3_with_randomly(test_path):
    test_path.makepyfile(
        test_one="""
        flag = True
        def test_a(): pass
        def test_b(): global flag; flag = False
        def test_c(): pass
        def test_d(): pass
        def test_e(): assert flag
        def test_f(): pass
        def test_g(): pass
        """
    )

    result = test_path.runpytest("--find-dependencies")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 7 tests.",
        "Dependent tests:",
        "  test_one.py::test_e depends on test_one.py::test_b"
    ])


def test_two_dependencies(test_path):
    test_path.makepyfile(
        test_one="""
        flag = True
        def test_a(): pass
        def test_b(): assert flag
        def test_c(): pass
        def test_d(): assert flag
        def test_e(): global flag; flag = False
        def test_f(): pass
        """
    )

    result = test_path.runpytest("--find-dependencies", "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 6 tests.",
        "Executed 21 tests in 7 test runs.",
        "Dependent tests:",
        "  test_one.py::test_b depends on test_one.py::test_e",
        "  test_one.py::test_d depends on test_one.py::test_e"
    ])


def test_single_dependency_in_other_module1(test_path):
    test_path.makepyfile(
        test_one="""
        import util
        def test_a(): pass
        def test_b(): util.set_flag(False)
        def test_c(): pass
        def test_d(): pass
        def test_e(): assert util.flag
        def test_f(): pass
        def test_g(): pass
        """
    )
    test_path.makepyfile(
        util="""
        flag = True
        def set_flag(new_flag):
            global flag
            flag = new_flag
        """
    )

    result = test_path.runpytest("--find-dependencies", "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 7 tests.",
        "Executed 19 tests in 4 test runs.",
        "Dependent tests:",
        "  test_one.py::test_e depends on test_one.py::test_b"
    ])


def test_single_dependency_in_other_module2(test_path):
    test_path.makepyfile(
        test_one="""
        import util
        def test_a(): pass
        def test_b(): assert util.flag
        def test_c(): pass
        def test_d(): pass
        def test_e(): util.set_flag(False)
        def test_f(): pass
        def test_g(): pass
        """
    )
    test_path.makepyfile(
        util="""
        flag = True
        def set_flag(new_flag):
            global flag
            flag = new_flag
        """
    )
    result = test_path.runpytest("--find-dependencies", "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 7 tests.",
        "Executed 20 tests in 5 test runs.",
        "Dependent tests:",
        "  test_one.py::test_b depends on test_one.py::test_e"
    ])


def test_permanent_dependency(test_path):
    test_path.makepyfile(
        test_one="""
        import util
        def test_a(): pass
        def test_b(): assert not util.lock_exists()
        def test_c(): pass
        def test_d(): util.create_lock()
        def test_e(): pass
        """
    )
    test_path.makepyfile(
        util="""
        import os
        def create_lock():
            with open("lock.lck", "w") as f:
                f.write("test")

        def lock_exists():
            return os.path.exists("lock.lck")
        """
    )
    result = test_path.runpytest("--find-dependencies", "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 5 tests.",
        "Executed 11 tests in 3 test runs.",
        "Tests failing permanently after all tests have run:",
        "  test_one.py::test_b"
    ])


def test_permanent_dependency_reversed_first(test_path):
    test_path.makepyfile(
        test_one="""
        import util
        def test_a(): pass
        def test_b(): assert not util.lock_exists()
        def test_c(): pass
        def test_d(): util.create_lock()
        def test_e(): pass
        """
    )
    test_path.makepyfile(
        util="""
        import os
        def create_lock():
            with open("lock1.lck", "w") as f:
                f.write("test")

        def lock_exists():
            return os.path.exists("lock1.lck")
        """
    )
    result = test_path.runpytest("--find-dependencies", "--reversed-first",
                                 "-p", "no:randomly")
    result.stdout.fnmatch_lines([
        "Run dependency analysis for 5 tests.",
        "Executed 10 tests in 2 test runs.",
        "The following tests are always failing and are "
        "excluded from the analysis:",
        "  test_one.py::test_b",
        "No dependent tests found."
    ])



from multiprocessing import Process

import pytest

CACHE_KEY_IDS = "find_dependencies/node_ids"
CACHE_KEY_FAILED = "find_dependencies/failed_ids"


class DependencyFinder:
    """Tries to find dependencies between tests using reverse executing
    first and more test executions using binary search to get dependent
    tests."""

    def __init__(self, session):
        self.session = session
        self.dependent_items = {}
        self.permanently_failed_items = []
        self.test_runs = 0
        self.test_number = 0

    def find_dependencies(self):
        reversed_first = self.session.config.getoption("reversed_first")
        reversed_items = self.session.items[::-1]
        if reversed_first:
            items1, items2 = reversed_items, self.session.items
        else:
            items1, items2 = self.session.items, reversed_items
        failed1 = self.run_tests(items1)
        failed2 = self.run_tests(items2)
        always_failing_items = failed1.intersection(failed2)

        # tests failing in both runs are not considered
        failed1, failed2 = failed1 - failed2, failed2 - failed1
        for item in sorted(failed1, key=str):
            self.check_failed_item(item, items1)

        # tests failing in the second run have to checked if they fail
        # permanently afterwards - in this case the we don't try to
        # find the dependency
        for item in sorted(failed2, key=str):
            self.check_failed_item(item, items2, check_permanent=True)

        print("\n=====================================================")
        print(f"Run dependency analysis for {len(items1)} tests.")
        print(f"Executed {self.test_number} tests in "
              f"{self.test_runs} test runs.")
        if always_failing_items:
            print("The following tests are always failing and "
                  "are excluded from the analysis:")
            for item in sorted(always_failing_items):
                print(f"  {item.nodeid}")

        if not self.dependent_items and not self.permanently_failed_items:
            print("No dependent tests found.")
        else:
            if self.permanently_failed_items:
                print("Tests failing permanently after all tests have run:")
                for item in self.permanently_failed_items:
                    print(f"  {item.nodeid}")
            if self.dependent_items:
                print("Dependent tests:")
                for item, dependent in self.dependent_items.items():
                    print(f"  {item.nodeid} depends on {dependent.nodeid}")
        print("=====================================================")

    def check_failed_item(self, item, items, failed=True,
                          check_permanent=False):
        if check_permanent and item in self.run_tests([item]):
            self.permanently_failed_items.append(item)
            return
        index = items.index(item)
        if failed:
            if index == 1:
                self.dependent_items[item] = items[0]
                return
            mid_index = index // 2
            sub_items_to_run = items[:mid_index] + [items[index]]
            sub_items = sub_items_to_run + items[mid_index:index]
        else:
            if index == len(items) - 2:
                self.dependent_items[item] = items[index + 1]
                return
            mid_index = index + 1 + (len(items) - index - 1) // 2
            sub_items_to_run = items[index + 1:mid_index] + [items[index]]
            sub_items = sub_items_to_run + items[mid_index:]

        failed_items = self.run_tests(sub_items_to_run)
        self.check_failed_item(item, sub_items, item in failed_items)

    def run_tests(self, items):
        items = {item.nodeid: item for item in items}
        self.session.config.cache.set(CACHE_KEY_IDS, list(items.keys()))
        p = Process(target=pytest.main,
                    args=[["-q", "--find-dependencies-internal"]])
        p.start()
        p.join()
        failed_node_ids = self.session.config.cache.get(CACHE_KEY_FAILED, [])
        self.test_runs += 1
        self.test_number += len(items)
        return set(items[key] for key in items if key in failed_node_ids)


def run_tests(session):
    all_items = {item.nodeid: item for item in session.items}
    node_ids = session.config.cache.get(CACHE_KEY_IDS, [])
    items = [all_items[node_id] for node_id in node_ids]
    failed_node_ids = []
    for index, item in enumerate(items):
        test_failed = session.testsfailed
        next_item = items[index + 1] if index + 1 < len(items) else None
        item.config.hook.pytest_runtest_protocol(item=item,
                                                 nextitem=next_item)
        if session.testsfailed > test_failed:
            failed_node_ids.append(item.nodeid)
    session.config.cache.set(CACHE_KEY_FAILED, failed_node_ids)



#!/usr/bin/env python
import os

from setuptools import setup, find_packages
from src.find_dependencies import __version__


def read(file_name):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, file_name)) as f:
        return f.read()


setup(
    name='pytest-find-dependencies',
    version=__version__,
    author='MrBean Bremen',
    author_email='hansemrbean@googlemail.com',
    maintainer='MrBean Bremen',
    maintainer_email='hansemrbean@googlemail.com',
    license='MIT',
    url='https://github.com/mrbean-bremen/pytest-find-dependencies',
    description='A pytest plugin to find dependencies between tests',
    long_description=read('README.rst'),
    py_modules=['pytest_find_dependencies'],
    python_requires='>=3.5',
    install_requires=['pytest>=3.5.0'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        "Programming Language :: Python :: 3.9",
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    entry_points={
        "pytest11": ["find_dependencies = find_dependencies.plugin"]
    }
)
